import { getArgumentValues, isInterfaceType, isObjectType, isUnionType, Kind, parse, print, } from 'graphql';
import { flow, isString } from 'lodash-es';
const directives = {};
export const registerDirectiveImplementation = (name, resolver) => {
    directives[name] = resolver;
};
/**
 * Extract Object types with @sourceFrom directives.
 *
 * Returns a map of type names to source package and function name tuples.
 */
export function extractSourceMapping(schema) {
    const sources = {};
    const sourceFrom = schema.getDirective('sourceFrom');
    if (!sourceFrom) {
        return sources;
    }
    for (const type of Object.values(schema.getTypeMap())) {
        if (isObjectType(type)) {
            for (const directive of type.astNode?.directives || []) {
                if (directive.name.value === 'sourceFrom') {
                    const values = getArgumentValues(sourceFrom, directive);
                    sources[type.name] = values.fn;
                    values['fn'];
                }
            }
        }
    }
    return sources;
}
/**
 * Extract fields types with @resolveBy directives.
 *
 * Returns a nested map of type names, field names, package and function name tuples.
 */
export function extractResolverMapping(schema) {
    const resolvers = {};
    Object.values(schema.getTypeMap()).forEach((type) => {
        if (isObjectType(type)) {
            Object.values(type.getFields()).forEach((field) => {
                field.astNode?.directives?.forEach((dir) => {
                    if (dir.name.value === 'resolveBy' || directives[dir.name.value]) {
                        const directive = schema.getDirective(dir.name.value);
                        if (directive) {
                            if (!resolvers[type.name]) {
                                resolvers[type.name] = {};
                            }
                            if (!resolvers[type.name][field.name]) {
                                resolvers[type.name][field.name] = [];
                            }
                            resolvers[type.name][field.name].push([
                                directive.name,
                                getArgumentValues(directive, dir),
                            ]);
                        }
                    }
                });
            });
        }
    });
    return resolvers;
}
/**
 * Extract a list of all union types.
 */
export function extractUnions(schema) {
    return Object.values(schema.getTypeMap())
        .filter(isUnionType)
        .map((type) => type.name);
}
/**
 * Extract a list of all interface types.
 */
export function extractInterfaces(schema) {
    return Object.values(schema.getTypeMap())
        .filter(isInterfaceType)
        .map((type) => type.name);
}
const loaded = {};
/**
 * Dynamically retrieve a function from a package.
 */
export async function loadFunction(spec) {
    if (loaded[spec]) {
        return loaded[spec];
    }
    const [module, fn] = spec.split('#');
    try {
        const m = await import(module);
        if (!m[fn]) {
            throw `Module "${module} has no function "${fn}".`;
        }
        loaded[spec] = m[fn];
        return m[fn];
    }
    catch (exc) {
        throw `Module "${module}" does not exist.`;
    }
}
/**
 * Clean up a schema string and remove everything that could confuse Gatsby.
 */
export function cleanSchema(schema) {
    const ast = parse(schema);
    const result = [];
    ast.definitions.forEach((def) => {
        if (def.kind === Kind.SCHEMA_DEFINITION ||
            def.kind === Kind.DIRECTIVE_DEFINITION) {
            return;
        }
        result.push(print(def));
    });
    return result.join('\n');
}
export function processDirectiveArguments(parent, args, spec) {
    return Object.fromEntries(Object.keys(spec).map((key) => {
        const val = spec[key];
        if (isString(val)) {
            if (val === '$') {
                return [key, parent];
            }
            if (val.match(/^\$.+/)) {
                return [key, args[val.substr(1)]];
            }
        }
        return [key, spec[key]];
    }));
}
export async function buildResolver(config) {
    const resolvers = {};
    for (const [name, spec] of config) {
        if (name === 'resolveBy') {
            resolvers[spec['fn']] = await loadFunction(spec['fn']);
        }
    }
    return ((source, args, context, info) => {
        const fns = [
            (parent) => {
                return parent?.[info?.fieldName];
            },
            ...config.map(([name, spec]) => {
                return (parent) => {
                    if (name === 'resolveBy') {
                        return resolvers[spec['fn']](parent, args, context, info);
                    }
                    return directives[name](parent, processDirectiveArguments(parent, args, spec), context, info);
                };
            }),
        ];
        const chain = flow(fns);
        return chain(source);
    });
}
